import hmac
from base64 import b64decode, b64encode
from hashlib import sha256
from typing import List, Optional, Set, Tuple

from bech32 import bech32_decode, bech32_encode, convertbits
from bip32 import BIP32
from Cryptodome import Random
from Cryptodome.Cipher import AES
from ecdsa import SECP256k1, SigningKey, VerifyingKey
from ecdsa.util import sigdecode_der, sigencode_der

from .exceptions import InvalidLnurl, InvalidUrl

LUD13_PHRASE = (
    "DO NOT EVER SIGN THIS TEXT WITH YOUR PRIVATE KEYS! IT IS ONLY USED "
    "FOR DERIVATION OF LNURL-AUTH HASHING-KEY, DISCLOSING ITS SIGNATURE "
    "WILL COMPROMISE YOUR LNURL-AUTH IDENTITY AND MAY LEAD TO LOSS OF FUNDS!"
)


def aes_decrypt(preimage: bytes, ciphertext_base64: str, iv_base64: str) -> str:
    """
    Decrypt a message using AES-CBC. LUD-10
    LUD-10, used in PayRequest success actions.
    """
    if len(preimage) != 32:
        raise ValueError("AES key must be 32 bytes long")
    if len(iv_base64) != 24:
        raise ValueError("IV must be 24 bytes long")
    cipher = AES.new(preimage, AES.MODE_CBC, b64decode(iv_base64))
    decrypted = cipher.decrypt(b64decode(ciphertext_base64))
    size = len(decrypted)
    pad = decrypted[size - 1]
    if (0 > pad > 16) or (pad > 1 and decrypted[size - 2] != pad):
        raise ValueError("Decryption failed. Error with padding.")
    decrypted = decrypted[: size - pad]
    if len(decrypted) == 0:
        raise ValueError("Decryption failed. Empty message.")
    try:
        return decrypted.decode("utf-8")
    except UnicodeDecodeError as exc:
        raise ValueError("Decryption failed. UnicodeDecodeError") from exc


def aes_encrypt(preimage: bytes, message: str) -> tuple[str, str]:
    """
    Encrypt a message using AES-CBC with a random IV.
    LUD-10, used in PayRequest success actions.
    """
    if len(preimage) != 32:
        raise ValueError("AES key must be 32 bytes long")
    if len(message) == 0:
        raise ValueError("Message must not be empty")
    iv = Random.get_random_bytes(16)
    cipher = AES.new(preimage, AES.MODE_CBC, iv)
    pad = 16 - len(message) % 16
    message += chr(pad) * pad
    ciphertext = cipher.encrypt(message.encode("utf-8"))
    return b64encode(ciphertext).decode(), b64encode(iv).decode("utf-8")


# LUD-04: auth base spec.
def lnurlauth_signature(k1: str, linking_key: bytes) -> tuple[str, str]:
    """
    Sign a k1 with a linking_key from (bip32 or signMessage) and a domain.

    Obtain the linking_key from lnurlauth_derive_linking_key or lnurlauth_derive_linking_key_sign_message.
    """
    auth_key = SigningKey.from_string(linking_key, curve=SECP256k1, hashfunc=sha256)
    sig = auth_key.sign_digest_deterministic(bytes.fromhex(k1), sigencode=sigencode_der)
    if not auth_key.verifying_key:
        raise ValueError("LNURLauth verifying_key does not exist")
    key = auth_key.verifying_key.to_string("compressed")
    return key.hex(), sig.hex()


def lnurlauth_verify(k1: str, key: str, sig: str) -> bool:
    """
    Verify a k1 with a key and a signature.
    """
    try:
        verifying_key = VerifyingKey.from_string(bytes.fromhex(key), hashfunc=sha256, curve=SECP256k1)
        verifying_key.verify_digest(bytes.fromhex(sig), bytes.fromhex(k1), sigdecode=sigdecode_der)
        return True
    except Exception as exc:
        print(exc)
        return False


# LUD-05: BIP32-based seed generation for auth protocol.
def lnurlauth_derive_linking_key(seed: str, domain: str) -> tuple[bytes, bytes]:
    """
    Derive a key from a masterkey.
    RETURN (linking_key, linking_key_pub) in hex
    """
    master_key = lnurlauth_master_key_from_seed(seed)
    hashing_key = BIP32.get_privkey_from_path(master_key, "m/138'/0")
    _path_suffix = lnurlauth_derive_path(hashing_key, domain)
    linking_key = BIP32.get_privkey_from_path(master_key, _path_suffix)
    linking_key_pub = BIP32.get_pubkey_from_path(master_key, _path_suffix)
    return linking_key, linking_key_pub


def lnurlauth_master_key_from_seed(seed: str) -> BIP32:
    """
    Derive a masterkey from a seed.
    RETURN (linking_key, linking_key_pub) in hex
    """
    master_key = BIP32.from_seed(bytes.fromhex(seed))
    assert master_key.privkey
    return master_key


def lnurlauth_derive_path(hashing_private_key: bytes, domain_name: str) -> str:
    """
    Derive a path suffix from a hashing_key.

    Take the first 16 bytes of the hash turn it into 4 longs and make a new derivation path with it.
    m/138'/<long1>/<long2>/<long3>/<long4>
    """
    derivation_material = hmac.digest(hashing_private_key, domain_name.encode(), "sha256")
    _path_suffix_longs = [int.from_bytes(derivation_material[i : i + 4], "big") for i in range(0, 16, 4)]
    _path_suffix = "m/138'/" + "/".join(str(i) for i in _path_suffix_longs)
    return _path_suffix


# LUD-13: signMessage-based seed generation for auth protocol.
def lnurlauth_message_to_sign() -> bytes:
    """
    Generate a message to sign for signMessage.
    """
    return sha256(LUD13_PHRASE.encode()).digest()


def lnurlauth_derive_linking_key_sign_message(domain: str, sig: bytes) -> tuple[bytes, bytes]:
    """
    Derive a key from a from signMessage from a node.
    param `sig` is a RFC6979 deterministic signature.
    """
    hashing_key = SigningKey.from_string(sha256(sig).digest(), curve=SECP256k1, hashfunc=sha256)
    linking_key = SigningKey.from_string(
        hmac.digest(hashing_key.to_string(), domain.encode(), "sha256"), curve=SECP256k1, hashfunc=sha256
    )
    assert linking_key.privkey and linking_key.verifying_key
    pubkey = linking_key.verifying_key.to_string("compressed")
    return linking_key.privkey, pubkey


def _bech32_decode(bech32: str, *, allowed_hrp: Optional[Set[str]] = None) -> Tuple[str, List[int]]:
    hrp, data = bech32_decode(bech32)

    if not hrp or not data or (allowed_hrp and hrp not in allowed_hrp):
        raise ValueError(f"Invalid data or Human Readable Prefix (HRP): {hrp}.")

    return hrp, data


def _lnurl_clean(lnurl: str) -> str:
    lnurl = lnurl.strip()
    return lnurl.replace("lightning:", "") if lnurl.startswith("lightning:") else lnurl


def url_decode(lnurl: str) -> str:
    """
    Decode a LNURL and return a url string without performing any validation on it.
    Use `lnurl.decode()` for validation and to get `Url` object.
    """
    _, data = _bech32_decode(_lnurl_clean(lnurl), allowed_hrp={"lnurl"})

    try:
        bech32_data = convertbits(data, 5, 8, False)
        assert bech32_data
        url = bytes(bech32_data).decode("utf-8")
        return url
    except UnicodeDecodeError:
        raise InvalidLnurl


def url_encode(url: str) -> str:
    """
    Encode a URL without validating it first and return a bech32 LNURL string.
    Use `lnurl.encode()` for validation and to get a `Lnurl` object.
    """
    try:
        bech32_data = convertbits(url.encode("utf-8"), 8, 5, True)
        assert bech32_data
        lnurl = bech32_encode("lnurl", bech32_data)
    except UnicodeEncodeError:
        raise InvalidUrl

    return lnurl.upper()



from __future__ import annotations

import json
import os
import re
from enum import Enum
from hashlib import sha256
from typing import Generator, Optional
from urllib.parse import parse_qs

from pydantic import (
    AnyUrl,
    ConstrainedStr,
    Json,
    ValidationError,
    parse_obj_as,
    validator,
)
from pydantic.validators import str_validator

from .exceptions import InvalidLnurlPayMetadata, InvalidUrl, LnAddressError
from .helpers import _bech32_decode, _lnurl_clean, url_decode, url_encode


class ReprMixin:
    def __repr__(self) -> str:
        attrs = [  # type: ignore
            outer_slot  # type: ignore
            for outer_slot in [slot for slot in self.__slots__ if not slot.startswith("_")]  # type: ignore
            if getattr(self, outer_slot)  # type: ignore
        ]  # type: ignore
        extra = ", " + ", ".join(f"{n}={getattr(self, n).__repr__()}" for n in attrs) if attrs else ""
        return f"{self.__class__.__name__}({super().__repr__()}{extra})"


class Bech32(ReprMixin, str):
    """Bech32 string."""

    __slots__ = ("hrp", "data")

    def __new__(cls, bech32: str, **_) -> "Bech32":
        return str.__new__(cls, bech32)

    def __init__(self, bech32: str, *, hrp: Optional[str] = None, data: Optional[list[int]] = None):
        str.__init__(bech32)
        self.hrp, self.data = (hrp, data) if hrp and data else self.__get_data__(bech32)

    @classmethod
    def __get_data__(cls, bech32: str) -> tuple[str, list[int]]:
        return _bech32_decode(bech32)

    @classmethod
    def __get_validators__(cls):
        yield str_validator
        yield cls.validate

    @classmethod
    def validate(cls, value: str) -> "Bech32":
        hrp, data = cls.__get_data__(value)
        return cls(value, hrp=hrp, data=data)


def ctrl_characters_validator(value: str) -> str:
    """Checks for control characters (unicode blocks C0 and C1, plus DEL)."""
    if re.compile(r"[\u0000-\u001f\u007f-\u009f]").search(value):
        raise InvalidUrl("URL contains control characters.")
    return value


def strict_rfc3986_validator(value: str) -> str:
    """Checks for RFC3986 compliance."""
    if os.environ.get("LNURL_STRICT_RFC3986", "0") == "1":
        if re.compile(r"[^]a-zA-Z0-9._~:/?#[@!$&'()*+,;=-]").search(value):
            raise InvalidUrl("URL is not RFC3986 compliant.")
    return value


def valid_lnurl_host(url: str) -> AnyUrl:
    """Validates the host part of a URL."""
    _url = parse_obj_as(AnyUrl, url)
    if not _url.host:
        raise InvalidUrl("URL host is required.")
    if _url.scheme == "http":
        if _url.host not in ["127.0.0.1", "0.0.0.0", "localhost"] and not _url.host.endswith(".onion"):
            raise InvalidUrl("HTTP scheme is only allowed for localhost or onion addresses.")
    return _url


class Url(AnyUrl):
    max_length = 2047  # https://stackoverflow.com/questions/417142/

    # LUD-17: Protocol schemes and raw (non bech32-encoded) URLs.
    allowed_schemes = {"https", "http", "lnurlc", "lnurlw", "lnurlp", "keyauth"}

    @property
    def is_lud17(self) -> bool:
        uris = ["lnurlc", "lnurlw", "lnurlp", "keyauth"]
        return any(self.scheme == uri for uri in uris)

    @classmethod
    def __get_validators__(cls) -> Generator:
        yield ctrl_characters_validator
        yield strict_rfc3986_validator
        yield valid_lnurl_host
        yield cls.validate

    @property
    def query_params(self) -> dict:
        return {k: v[0] for k, v in parse_qs(self.query).items()}


class CallbackUrl(Url):
    """URL for callbacks. exclude lud17 schemes."""

    allowed_schemes = {"https", "http"}


class LightningInvoice(Bech32):
    """Bech32 Lightning invoice."""


class LightningNodeUri(ReprMixin, str):
    """Remote node address of form `node_key@ip_address:port_number`."""

    __slots__ = ("key", "ip", "port")

    def __new__(cls, uri: str, **_) -> "LightningNodeUri":
        return str.__new__(cls, uri)

    def __init__(self, uri: str, *, key: Optional[str] = None, ip: Optional[str] = None, port: Optional[str] = None):
        str.__init__(uri)
        self.key = key
        self.ip = ip
        self.port = port

    @classmethod
    def __get_validators__(cls):
        yield str_validator
        yield cls.validate

    @classmethod
    def validate(cls, value: str) -> "LightningNodeUri":
        try:
            key, netloc = value.split("@")
            ip, port = netloc.split(":")
        except Exception:
            raise ValueError

        return cls(value, key=key, ip=ip, port=port)


class Lnurl(ReprMixin, str):
    url: Url
    lud17_prefix: Optional[str] = None

    def __new__(cls, lightning: str) -> Lnurl:
        url = cls.clean(lightning)
        _url = url.replace(url.scheme, "https")
        return str.__new__(cls, _url)

    def __init__(self, lightning: str):
        url = self.clean(lightning)
        if url.is_lud17:
            self.lud17_prefix = url.scheme
            _url = parse_obj_as(Url, url.replace(url.scheme, "https"))
            self.url = _url
            return str.__init__(_url)
        self.url = url
        return str.__init__(url)

    @classmethod
    def __get_validators__(cls):
        yield str_validator
        yield cls.validate

    @classmethod
    def clean(cls, lightning: str) -> Url:
        lightning = _lnurl_clean(lightning)
        if lightning.lower().startswith("lnurl1"):
            url = parse_obj_as(Url, url_decode(lightning))
            return url
        url = parse_obj_as(Url, lightning)
        return url

    @classmethod
    def validate(cls, lightning: str) -> Lnurl:
        _ = cls.clean(lightning)
        return cls(lightning)

    @property
    def bech32(self) -> Optional[Bech32]:
        """Returns Bech32 representation of the Lnurl if it is a Bech32 encoded URL."""
        return parse_obj_as(Bech32, url_encode(self.url))

    # LUD-04: auth base spec.
    @property
    def is_login(self) -> bool:
        return self.url.query_params.get("tag") == "login"

    # LUD-08: Fast withdrawRequest.
    @property
    def is_fast_withdraw(self) -> bool:
        q = self.url.query_params
        return (
            q.get("tag") == "withdrawRequest"
            and q.get("k1") is not None
            and q.get("minWithdrawable") is not None
            and q.get("maxWithdrawable") is not None
            and q.get("defaultDescription") is not None
            and q.get("callback") is not None
        )

    # LUD-17: Protocol schemes and raw (non bech32-encoded) URLs.
    @property
    def is_lud17(self) -> bool:
        return self.lud17_prefix is not None

    @property
    def lud17(self) -> Optional[str]:
        if not self.lud17_prefix:
            return None
        url = self.url.replace(self.url.scheme, self.lud17_prefix)
        return url


class LnAddress(ReprMixin, str):
    """Lightning address of form `user+tag@host`"""

    slots = ("address", "url", "tag")

    def __new__(cls, address: str) -> LnAddress:
        return str.__new__(cls, address)

    def __init__(self, address: str):
        str.__init__(address)
        if not self.is_valid_lnaddress(address):
            raise LnAddressError("Invalid Lightning address.")
        self.url = self.__get_url__(address)
        if "+" in address:
            self.tag: Optional[str] = address.split("+", 1)[1].split("@", 1)[0]
            self.address = address.replace(f"+{self.tag}", "", 1)
        else:
            self.tag = None
            self.address = address

    # LUD-16: Paying to static internet identifiers.
    @validator("address")
    def is_valid_lnaddress(cls, address: str) -> bool:
        # A user can then type these on a WALLET. The <username> is limited
        # to a-z-1-9-_.. Please note that this is way more strict than common
        # email addresses as it allows fewer symbols and only lowercase characters.
        lnaddress_regex = r"[a-z0-9\._%+-]+@[A-Za-z0-9\.-]+\.[A-Za-z]{2,63}"
        return re.fullmatch(lnaddress_regex, address) is not None

    @classmethod
    def __get_url__(cls, address: str) -> CallbackUrl:
        name, domain = address.split("@")
        url = ("http://" if domain.endswith(".onion") else "https://") + domain + "/.well-known/lnurlp/" + name
        return parse_obj_as(CallbackUrl, url)


class LnurlPayMetadata(ReprMixin, str):
    # LUD-16: Paying to static internet identifiers. "text/identifier", "text/email", "text/tag"
    # LUD-20: Long payment description for pay protocol. "text/long-desc"
    valid_metadata_mime_types = {
        "text/plain",
        "image/png;base64",
        "image/jpeg;base64",
        "text/identifier",
        "text/email",
        "text/tag",
        "text/long-desc",
    }

    __slots__ = ("_list",)

    def __new__(cls, json_str: str, **_) -> LnurlPayMetadata:
        return str.__new__(cls, json_str)

    def __init__(self, json_str: str, *, json_obj: Optional[list] = None):
        str.__init__(json_str)
        self._list = json_obj if json_obj else self.__validate_metadata__(json_str)

    @classmethod
    def __validate_metadata__(cls, json_str: str) -> list[tuple[str, str]]:
        try:
            parse_obj_as(Json[list[tuple[str, str]]], json_str)
            data = [(str(item[0]), str(item[1])) for item in json.loads(json_str)]
        except ValidationError:
            raise InvalidLnurlPayMetadata

        clean_data = [x for x in data if x[0] in cls.valid_metadata_mime_types]
        mime_types = [x[0] for x in clean_data]
        counts = {x: mime_types.count(x) for x in mime_types}

        if (
            not clean_data
            or ("text/plain" not in mime_types and "text/long-desc" not in mime_types)
            or counts["text/plain"] > 1
        ):
            raise InvalidLnurlPayMetadata

        return clean_data

    @classmethod
    def __get_validators__(cls):
        yield str_validator
        yield cls.validate

    @classmethod
    def validate(cls, value: str) -> LnurlPayMetadata:
        return cls(value, json_obj=cls.__validate_metadata__(value))

    @property
    def h(self) -> str:
        return sha256(self.encode("utf-8")).hexdigest()

    @property
    def text(self) -> str:
        output = ""

        for metadata in self._list:
            if metadata[0] == "text/plain":
                output = metadata[1]
                break

        return output

    @property
    def images(self) -> list[tuple[str, str]]:
        return [x for x in self._list if x[0].startswith("image/")]

    def list(self) -> list[tuple[str, str]]:
        return self._list


class InitializationVectorBase64(ConstrainedStr):
    min_length = 24
    max_length = 24


class CiphertextBase64(ConstrainedStr):
    min_length = 24
    max_length = 4096


class Max144Str(ConstrainedStr):
    max_length = 144


# LUD-04: auth base spec.
class LnurlAuthActions(Enum):
    """Enum for auth actions"""

    login = "login"
    register = "register"
    link = "link"
    auth = "auth"


class LnurlPaySuccessActionTag(Enum):
    """Enum for success action tags"""

    aes = "aes"
    message = "message"
    url = "url"


class LnurlStatus(Enum):
    """Enum for status"""

    ok = "OK"
    error = "ERROR"


class LnurlResponseTag(Enum):
    """Enum for response tags"""

    login = "login"
    channelRequest = "channelRequest"
    hostedChannelRequest = "hostedChannelRequest"
    payRequest = "payRequest"
    withdrawRequest = "withdrawRequest"
